package kernel.net;

import kernel.log.Log;
import kernel.peripherals.Peripherals.Peripheral;
import kernel.net.INetworkInterface;
import kernel.net.Package.NetworkID;

using tink.CoreApi;

@:structInit typedef Route = {
	interf:INetworkInterface,
	rep:NetworkID,
	cost:Int
}

/**
	Manages network routing.
**/
class Routing {
	/**
		Depends on: Peripheral
	**/
	public static inline final UPDATE_WAIT_TIME:Float = 1;

	public static var onNewNeigbor(default, null):Signal<Int>;

	private static final onNewNeigborTrigger:SignalTrigger<NetworkID> = Signal.trigger();
	private static final routingTable:Map<NetworkID, Route> = new Map();
	private static var routeUpdateInQueue:Bool = false;

	@:allow(kernel.Init)
	private static function init() {
		onNewNeigbor = onNewNeigborTrigger.asSignal();
	}

	@:allow(kernel.net.Net)
	private static function setup() {
		routingTable.set(Net.networkID, {interf: Loopback.instance, cost: 0, rep: Net.networkID});
		brodcastRoutingTable();
	}

	/**
		Prepares an brodcast of the current routing table. If a brodcast is already in queue it will be ignored.
		After UPDATE_WAIT_TIME seconds the routing table will be brodcasted. This is done to prevent spamming the network.
	**/
	private static function prepareRouteUpdate() {
		if (routeUpdateInQueue) {
			return;
		}

		routeUpdateInQueue = true;

		new Timer(UPDATE_WAIT_TIME, () -> {
			brodcastRoutingTable();
			routeUpdateInQueue = false;
		});
	}

	/**
		Brodcast the current routing table to all peers.
	**/
	private static function brodcastRoutingTable() {
		var pack = newRoutDiscoverPackage();

		for (modem in Peripheral.getAllModems()) {
			modem.send(Net.BRODCAST_PORT, Net.networkID, pack);
		}
	}

	/**
		Handle incomming packages that involve route discovery.
	**/
	@:allow(kernel.net.Net)
	private static function handleRoutePackage(pack:Package<Noise>, interf:INetworkInterface):Void {
		addPossibleRoute(pack.fromID, interf, 0, pack.fromID);

		var shouldRespond:Bool = switch pack.type {
			case RouteDiscoverResponse(routes):
				for (route in routes) {
					addPossibleRoute(route.id, interf, route.cost, pack.fromID);
				}
				false;
			case RouteDiscover(routes):
				for (route in routes) {
					addPossibleRoute(route.id, interf, route.cost, pack.fromID);
				}
				true;
			case RouteDiscoverUpdate(routes):
				for (route in routes) {
					addPossibleRoute(route.id, interf, route.cost, pack.fromID);
				}
				false;
			default:
				Log.error("Expected package to be a Route discover package");
				false;
		};

		if (!shouldRespond) {
			return;
		}

		// Respond to peer
		var response:Package<Noise> = {
			toID: pack.fromID,
			fromID: Net.networkID,
			msgID: null,
			type: RouteDiscoverResponse(genRouteList()),
			data: null,
			ttl: 0, // Prevent forwarding
		}

		interf.send(response.toID, Net.networkID, response);
	}

	/**
		Generate a list of routes to send to a peer based on the current routing table.
	**/
	private static function genRouteList():Array<{id:NetworkID, cost:Int}> {
		var routes:Array<{id:NetworkID, cost:Int}> = [];
		for (k => v in routingTable) {
			routes.push({
				id: k,
				cost: v.cost
			});
		}
		return routes;
	}

	private static function newRoutDiscoverPackage():Package<Noise> {
		var pack:Package<Noise> = {
			type: RouteDiscover(genRouteList()),
			toID: Net.BRODCAST_PORT,
			msgID: null,
			fromID: Net.networkID,
			data: null,
			ttl: 0, // Prevent forwarding
		}

		return pack;
	}

	/**
		Called when a route to a client has been disoverd.
		Its possible to be called multiple times with the same id but different addr.
	**/
	private static function addPossibleRoute(toID:Int, interf:INetworkInterface, cost:Int, rep:NetworkID) {
		if (toID == Net.networkID) {
			return;
		}

		var fullCost = cost + interf.getBaseRoutingCost();

		if (routingTable.exists(toID)) {
			if (routingTable[toID].cost > fullCost) {
				routingTable[toID] = {interf: interf, cost: cost + interf.getBaseRoutingCost(), rep: rep};
				prepareRouteUpdate();
			}
		} else {
			routingTable[toID] = {interf: interf, cost: cost + interf.getBaseRoutingCost(), rep: rep};
			onNewNeigborTrigger.trigger(toID);
			prepareRouteUpdate();
		}
	}

	public static function getRouteToID(networkID:NetworkID):{interf:INetworkInterface, rep:NetworkID} {
		if (networkID == Net.networkID) {
			return {interf: Loopback.instance, rep: Net.networkID};
		}

		var route:Null<Route> = routingTable[networkID];

		if (route == null) {
			return null;
		} else {
			return {interf: route.interf, rep: route.rep};
		}
	}

	public static function getRouteTable():Map<NetworkID, Route> {
		return routingTable;
	}
}
