package lib;

import kernel.log.Log;
import kernel.binstore.BinStore;
import kernel.peripherals.Screen;
import kernel.peripherals.Peripherals.Peripheral;
import kernel.ps.Process;
import kernel.ps.ProcessManager;
import bin.KernelLog;
import lib.ui.elements.UIElement;
import lib.ui.elements.TextElement;
import lib.ui.elements.RootElement;
import kernel.KernelEvents;
import bin.Terminal;
import kernel.ui.WindowManager;
import kernel.ui.WindowContext;

/**
	The WindowContext that spawns all other contexts. This is the main terminal.
	Listens to global hotkey "POS1"/"HOME" to switch back to this context.
**/
class HomeContext {
	private static inline final MAX_CONTEXT:Int = 10;

	private var ctx:WindowContext = null;
	private final workspaces:Map<Int, WindowContext> = [];
	private var currentWorkspace:Int = -1;
	private var requestRender:Void->Void;
	private var renderer:RootElement;

	private var selectedOutput:String = "main";
	private var selectedOutputIndex:Int = -1;

	private final listedApps:Array<String> = ["terminal", "log", "pfclient"];

	public function new() {}

	public function run() {
		// Create main terminal context
		var stateless = WindowManager.createNewStatelessContext();
		ctx = stateless.ctx;
		requestRender = stateless.requestRender;
		WindowManager.focusContextToOutput(ctx, "main");

		renderer = new RootElement();
		renderer.setTitle("Home");

		ctx.delegateEvents(renderer);
		stateless.setRenderFunc(this.render);
		requestRender();

		// Register global key bindings to react to main terminal
		KernelEvents.onKey.handle(e -> {
			// Is HOME pressed?
			// TODO: remove magic number
			if (e.keyCode == 268) {
				focusMainTerm();
			}
		});
	}

	/**
		Adds a context to the list of contexts to manage. Returns the context ID. Returns -1 if the context could not be added.
	**/
	private function addContextNextWorkspace(context:WindowContext):Int {
		for (i in 0...MAX_CONTEXT) {
			if (!workspaces.exists(i)) {
				workspaces[i] = context;
				return i;
			}
		}

		return -1;
	}

	private function addContextToWorkspace(context:WindowContext, id:Int) {
		if (!workspaces.exists(id)) {
			workspaces[id] = context;
		}
	}

	private function focusContext(id:Int) {
		if (workspaces.exists(id)) {
			WindowManager.focusContextToOutput(workspaces[id], selectedOutput);
			currentWorkspace = id;
		}
	}

	private function focusMainTerm() {
		requestRender();
		WindowManager.focusContextToOutput(ctx, "main");
		currentWorkspace = -1;
	}

	private function handleSelectContext(contextId:Int) {
		if (workspaces[contextId] == null) {
			return;
		}

		focusContext(contextId);
	}

	private function spawnPs(binName:String) {
		var bin = BinStore.getBinByAlias(binName);

		if (bin == null) {
			Log.error('Could not find bin: ${binName}');
			return;
		}

		var ps = Type.createInstance(bin.c, []);
		var pid = ProcessManager.run(ps, {});
		var lastContextID = -1;

		for (ctx in WindowManager.getContextByPID(pid)) {
			lastContextID = addContextNextWorkspace(ctx);
		}

		if (lastContextID == -1) {
			return;
		}
		focusContext(lastContextID);

		if (selectedOutputIndex != -1) {
			requestRender();
		}
	}

	private function cycleOutput() {
		var screenAddr = Peripheral.findAddrByType(Screen.TYPE_NAME);

		if (selectedOutputIndex == -1) {
			selectedOutputIndex = 0;
			selectedOutput = screenAddr[selectedOutputIndex];
		} else if (selectedOutputIndex >= screenAddr.length - 1) {
			selectedOutputIndex = -1;
			selectedOutput = "main";
		} else {
			selectedOutputIndex++;
			selectedOutput = screenAddr[selectedOutputIndex];
		}

		requestRender();
	}

	private function render() {
		ctx.setCursorBlink(false);

		var workspaceIDs:Array<Int> = [for (k => v in workspaces) k];
		workspaceIDs.sort((a, b) -> a - b);

		var children:Array<UIElement> = [
			for (i in workspaceIDs)
				new TextElement('Switch to ${i + 1}', {uiEvents: {onClick: this.handleSelectContext.bind(i)}})
		];

		for (i in 0...listedApps.length) {
			children.push(new TextElement('Add ${BinStore.getNameByAlias(listedApps[i])}', {uiEvents: {onClick: this.spawnPs.bind(listedApps[i])}}));
		}

		children.push(new TextElement('Output: ${selectedOutput}', {uiEvents: {onClick: this.cycleOutput}}));
		children.push(new TextElement('Exit', {style: {bgColor: Red}, uiEvents: {onClick: KernelEvents.shutdown}}));

		renderer.setChildren(children);

		renderer.render(ctx.getSize()).renderToContext(ctx);
	}
}
